(module pgbgl
   (export
;;   For the sake of simplicity, the Scheme
;;   procedures here below are (more or less)
;;   in the same order as the corresponding
;;   C functions are documented in "PostgreSQL
;;   Programmer's Guide", chapter 16, "libpq".
;;    
;;   Database Connection Procedures
;;    
    (pgbgl:setdbLogin pghost
		      pgport
		      pgoptions
		      pgtty
		      dbName
		      login
		      pwd)
    (pgbgl:connectdb conninfo)
    (pgbgl:finish pg-conn)
    (pgbgl:db pg-conn)
    (pgbgl:status pg-conn)
    (pgbgl:status-not-ok? pg-conn)
    (pgbgl:errorMessage pg-conn)
;;
;;   Query Execution Procedures
;;    
    (pgbgl:exec pg-conn query)
    (pgbgl:resultStatus result)
    (pgbgl:result-ok? res)
    (pgbgl:result-empty? res)
;    (pgbgl:resultErrorMessage res)
    (pgbgl:ntuples res)
    (pgbgl:nfields res)
    (pgbgl:fnames result)
    (pgbgl:fname result i)
    (pgbgl:ftype result field-number)
    (pgbgl:getvalues result tuple)
    (pgbgl:getvalue result tuple i)
    (pgbgl:gettuple result tuple)
    (pgbgl:print fout res printOpt)
;;   for pgbgl:print:
    (pgbgl:set-print-options! header
			      align
			      standard
			      html3
			      expanded
			      pager
			      fieldSep
			      tableOpt
			      caption
			      fieldName)
    (pgbgl:FieldAccessor this-field
			 Opt)
    (pgbgl:getfieldSep OPt)
    (pgbgl:getfieldName OPt)
;;   it was for pgbgl:print...  
    (pgbgl:printTuples res
		       fout
		       printAttName
		       terseOutput
		       width)
    (pgbgl:clear result)
;;
;;   Tracing Procedures
;;    
    (pgbgl:trace pg-conn
		 the-output-port)
    (pgbgl:untrace pg-conn
		   the-output-port)
;;
;;    
    (pgbgl:make-table! conn
		       tablename
		       fields)
    (pgbgl:insert-record! conn
			  tablename
			  record)
    (pgbgl:for-each proc result)
    (pgbgl:tuples result)
    )
   (include "my-headers.inc") 
   (extern 
    (type sshort short "sshort") 
    (type sint short "sint")
    (include "my-headers.h") 
    (PQsetdbLogin::PGconn*
     (string string string string string string string)
     "PQsetdbLogin") 
    (PQconnectdb::PGconn* string "PQconnectdb")
    (PQfinish::void (PGconn*) "PQfinish") 
    (PQdb::string (PGconn*) "PQdb")
    (PQhost::string (PGconn*) "PQhost")  
    (PQtty::string (PGconn*) "PQtty")
    (PQstatus::ConnStatusType (PGconn*) "PQstatus")
    (PQerrorMessage::string (PGconn*) "PQerrorMessage")
    (PQexec::PGresult* (PGconn* string) "PQexec") 
    (PQresultStatus::ExecStatusType (PGresult*) "PQresultStatus") 
;    (PQresultErrorMessage::string (PGresult*)
;				  "PQresultErrorMessage")
    (PQntuples::int (PGresult*) "PQntuples")
    (PQnfields::int (PGresult*) "PQnfields")
    (PQfname::string (PGresult* int) "PQfname")
    (PQftype::Oid (PGresult* int) "PQftype")
    (PQgetvalue::string (PGresult* int int) "PQgetvalue")
    (PQprint::void (FILE PGresult* PQprintOpt*) "PQprint")
    (PQprintTuples::void (PGresult* FILE int int int) "PQprintTuples") 
    (PQclear::int (PGresult*) "PQclear") 
    (PQtrace::void (PGconn* FILE) "PQtrace") 
    (PQuntrace::void (PGconn* FILE) "PQuntrace") 
 ))

(define (pgbgl:setdbLogin   ;; returns pg-conn
	 pghost             ;; string
	 pgport             ;; string
	 pgoptions          ;; string
	 pgtty              ;; string (for debug)
	 dbName             ;; string
	 login              ;; string
	 pwd                ;; string
	 )
   (PQsetdbLogin pghost pgport pgoptions pgtty dbName login pwd))

(define (pgbgl:connectdb
	 conninfo)           ;; string
   (PQconnectdb conninfo))

(define (pgbgl:finish pg-conn)
   (PQfinish pg-conn)
   (pgbgl:status pg-conn))

(define (pgbgl:db pg-conn)   ;; returns dbName
   (PQdb pg-conn))

(define (pgbgl:status        ;; returns connexion status:
                             ;; ConnStatusType-CONNECTION_BAD
	                     ;; or ConnStatusType-CONNECTION_OK
	                     ;; to be compared by =ConnStatusType?
	 pg-conn)
   (PQstatus pg-conn))

(define (pgbgl:status-not-ok? pg-conn)
   (=ConnStatusType? (PQstatus pg-conn) 
		     (ConnStatusType-CONNECTION_BAD)))

(define (pgbgl:errorMessage pg-conn) ; 
   (PQerrorMessage pg-conn))

(define (pgbgl:exec pg-conn
		    query    ;; a string, SQL query
		    )        ;; returns the result
   (PQexec pg-conn query))

(define (pgbgl:resultStatus result)
   (PQresultStatus result))

(define (pgbgl:result-ok? res)
   (=ExecStatusType? (PQresultStatus res) 
		     (ExecStatusType-PGRES_COMMAND_OK)))

(define (pgbgl:result-empty? res)
   (=ExecStatusType? (PQresultStatus res) 
		     (ExecStatusType-PGRES_EMPTY_QUERY)))

;(define (pgbgl:resultErrorMessage res)
;   (PQresultErrorMessage res))

(define (pgbgl:ntuples res)  ;; returns the [integer]
   (PQntuples res))           ;; number of tuples in res

(define (pgbgl:nfields res)  ;; returns the [integer]
   (PQnfields res))          ;; number of fields in each
                             ;; tuple of res

(define (pgbgl:fnames result)
  (let ((nfields (pgbgl:nfields result)))
    (define (iter i)
      (if (= i nfields) '()
          (cons (pgbgl:fname result i) (iter (+ i 1)))))
    (iter 0)))

(define (pgbgl:fname result i)
   (PQfname result i))

(define (pgbgl:ftype result field-number)
   (PQftype result field-number))

(define (pgbgl:getvalues result tuple)
  (let ((nfields (pgbgl:nfields result)))
    (define (iter i)
      (if (= i nfields) '()
          (cons (pgbgl:getvalue result tuple i) (iter (+ i 1)))))
    (iter 0)))

(define (pgbgl:getvalue result tuple i)
   (PQgetvalue result tuple i))

(define (pgbgl:gettuple result tuple)
  (map (lambda (n v) (cons (string->symbol n) v))
       (pgbgl:fnames result)
       (pgbgl:getvalues result tuple)))

(define (pgbgl:print fout    ;; output stream
		     res     ;; result of pgbgl:exec
		     printOpt) ; print options
   (PQprint fout res printOpt)
   (pgbgl:resultStatus res))

(define (pgbgl:set-print-options!
	 ;; first 6 arguments are od type PQBOOL : integer as boolean
	 header              ;; integer, print output fiels headings
	                     ;; and row count
	 align               ;; integer, fill align the fields
	 standard            ;; integer, old brain dead format (sic)
	 html3               ;; integer, output html tables
	 expanded            ;; integer, expand tables
	 pager               ;; integer, use pager for output
	 fieldSep            ;; string, field separator
	 tableOpt            ;; string, insert to HTML <table ...>
	 caption             ;; string, HTML <caption>
	 fieldName)          ;; vector of strings, repl. field names
   (PQprintOPt* (integer->char header)
		(integer->char align)
		(integer->char standard)
		(integer->char html3)
		(integer->char expanded)
		(integer->char pager)
		fieldSep
		tableOpt
		caption
		(let* ((fieldName-size (vector-length fieldName))
		       (opt (make-string* fieldName-size)))
		   (do ((i 0 (+ i 1)))
			 ((= i fieldName-size)
			  (string*-set! opt  fieldName-size
					string-null))
		      (string*-set! opt i
				    (vector-ref fieldName i)))
		   opt)))

(define (pgbgl:FieldAccessor
	 this-field         ;; string, the name of the field
                            ;; to be selected
	 Opt)               ;; the struct
   ((class-field-accessor this-field) Opt))

(define (pgbgl:getfieldSep OPt)
   (PQprintOPt*-fieldSep OPt))

(define (pgbgl:getfieldName OPt)
   (let ((fieldName (PQprintOPt*-fieldName OPt)))
      (if (string*-null? fieldName)
	  "NULL"
	  (string*-ref fieldName 0))))

(define (pgbgl:printTuples
	 res                 ;; the result of pgbgl:exec
	 fout                ;; output stream
	 printAttName        ;; integer, 0 or 1
	 terseOutput         ;; integer, 0 or 1
	 width)              ;; integer
   (PQprintTuples res fout printAttName terseOutput width)
   (pgbgl:resultStatus res))

(define (pgbgl:clear result)
   (PQclear result))

(define (pgbgl:trace         ;; [de-] activate trace facility
	 pg-conn             ;; pg-conn
	 fout)               ;; output stream
   (PQtrace pg-conn fout)
   (pgbgl:status pg-conn))

(define (pgbgl:untrace         ;; [de-] activate trace facility
	 pg-conn             ;; pg-conn
	 fout)               ;; output stream
   (PQuntrace pg-conn fout)
   (pgbgl:status pg-conn))

(define (pgbgl:for-each proc result)
   (let ((ntuples (pgbgl:ntuples result)))
     (define (iter i)
       (cond ((= i ntuples) ntuples)
	     (else (proc (pgbgl:gettuple result i))
		   (iter (+ i 1)))))
     (iter 0)))

(define (pgbgl:tuples result)
  (let ((ntuples (pgbgl:ntuples result)))
    (define (iter i)
      (cond ((= i ntuples) '())
            (else (cons (pgbgl:gettuple result i)
                  (iter (+ i 1))))))
    (iter 0)))

(define (list->sqlcmd lst)
  (define (cmd->string cmd)
    (cond ((string? cmd) (sqlquote cmd))
          ((number? cmd) (number->string cmd))
          ((symbol? cmd) (symbol->string cmd))))
    (if (null? lst) ""
      (let ((cmd (car lst)))
        (string-append
          (cond ((pair? cmd) (string-append "(" (list->sqlcmd cmd) ")"))
                (else (string-append (cmd->string cmd))))
	  (let ((tail (cdr lst)))
	    (if (not (null? tail))
	        (string-append " " (list->sqlcmd tail)) ""))))))

(define (sqlquote str)
  (string-append "'" (sqlescape str) "'"))

(define (sqlescape str)
  (let* ((retlst '())
         (push (lambda (c) (set! retlst (cons c retlst)))))
    (for-each (lambda (chr)
                (if (or (eqv? chr #\\) (eqv? chr #\')) (push #\\) '())
                (push chr))
      (string->list str))
    (list->string (reverse retlst))))

(define (displaysql lst)
  (display (list->sqlcmd lst))
  (newline))
(define (as-string obj)
  (cond ((string? obj) obj)
        ((char? obj) (make-string 1 obj))
        ((symbol? obj) (symbol->string obj))
        ((number? obj) (number->string obj))
        ((procedure? obj) (as-string obj))
;;      ((procedure? obj) (as-string (obj)))
        ((pair? obj) (interpolate-list "(" " " ")" 
                       (lambda (x) (as-string x)) obj))
        (else "?as-string")))

(define (interpolate-list afore twixt aft proc lst)
  (define (iter lst)
    (let ((elt (car lst))
          (tail (cdr lst)))
       (string-append (proc elt)
         (if (not (null? tail))
           (string-append (as-string twixt) (iter tail))
           (as-string aft)))))
  (string-append (as-string afore) (iter lst)))

(define (pgbgl:make-table! conn tablename fields)
  (let ((sql-string (string-append
                      "CREATE TABLE " tablename " "
                      (interpolate-list "(" ", " ")" 
                        (lambda (p)
                          (string-append 
                            (as-string (car p)) " " (as-string (cadr p))))
                        fields))))
       (pgbgl:exec conn sql-string)))

(define (pgbgl:insert-record! conn tablename record)
   (let ((sql-string
	  (string-append
	   "INSERT INTO " tablename " "
	   (interpolate-list "(" ", " ")" 
			     (lambda (p)
				(string-append (as-string (car p)) " "))
			     record)
	   " VALUES "
	   (interpolate-list "(" ", " ")" 
			     (lambda (p)
				(string-append (as-string (cdr p)) " "))
			     record))))
      (pgbgl:exec conn sql-string)))








